/***************************************************************************
                          kschoolmenu.cpp  -  description
                             -------------------
    begin                : Sun Aug 11 22:14:36 CEST 2002
    copyright            : (C) 2002 by Morten Hustveit
    email                : mortehu@ifi.uio.no
    modified		 : April 2003 and later 
    email		 : faj@bzz.no
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "kschoolmenu.h"

#include <unistd.h>

#include <qdir.h>
#include <qlayout.h>
#include <qlabel.h>
#include <qcombobox.h>
#include <qpushbutton.h>
#include <qlistview.h>
#include <qsplitter.h>
#include <qheader.h>
#include <qtextstream.h>

#include <klocale.h>
#include <kglobal.h>
#include <kiconloader.h>
#include <kdesktopfile.h>
#include <klineeditdlg.h>
#include <kmessagebox.h>
#include <kstandarddirs.h>

#include <iostream>

MenuItem::MenuItem(const QPixmap& icon, const QString& name, const QString& path)
  : icon(icon),
    name(name),
    path(path)
{
  hidden = false;

  if(path.contains(".desktop"))
  {
    QFile file(path);

    if(file.open(IO_ReadOnly))
    {
      desktopFile = QTextStream(&file).read();
      file.close();
    }
  }
}

MenuItem::~MenuItem()
{
  for(uint i = 0; i < children.size(); ++i)
    delete children[i];
}

void MenuItem::Show()
{
  hidden = false;
}

void MenuItem::Hide()
{
  hidden = true ;
}

template<class Vector, class Item>
static void push_back(Vector& v, const Item* i)
{
  v.resize(v.size() + 1);
  v.insert(v.size() - 1, i);
}

static void read_desktop_files(MenuItem* item, const QString& path)
{
  if(item->path.contains(".desktop"))
  { 
    KDesktopFile desktopFile(path + item->path.right(item->path.length() - item->path.findRev('/')));
    // check if we should really hide the item
    if(desktopFile.readBoolEntry("NoDisplay") || desktopFile.readBoolEntry("Hidden"))
      item->Hide();
  }
  else
  {
    KDesktopFile desktopFile(path + item->path.right(item->path.length() - item->path.findRev('/')) + "/.directory");
    // check if we should really hide the item
    if(desktopFile.readBoolEntry("NoDisplay") || desktopFile.readBoolEntry("Hidden"))
      item->Hide();
    
    // call read_desktop_files() for all files and subdirectories in path
    for(uint i = 0; i < item->children.size(); ++i) 
      read_desktop_files(item->children[i], path + item->path.right(item->path.length() - item->path.findRev('/'))); 
  }
}

static void read_dir(const QString& path, MenuItem& parent)
{
  QDir dir(path);
  dir.setSorting(QDir::Name);
  dir.setNameFilter("*.desktop");
  dir.setMatchAllDirs(true);

  for(uint i = 0; i < dir.count(); ++i)
  {
    if(dir[i] == "." || dir[i] == "..")
      continue;

    QFileInfo info(dir, dir[i]);

    if(info.isDir())
    {
      KDesktopFile desktopFile(info.filePath() + "/.directory");
      QString itemName;

      if(desktopFile.readBoolEntry("NoDisplay") || desktopFile.readBoolEntry("Hidden"))
        continue;

      itemName = desktopFile.readName();
      if (!itemName)
        itemName = dir[i];
      
      MenuItem* item = new MenuItem(
        KGlobal::iconLoader()->loadIcon(desktopFile.readIcon(), KIcon::Desktop, KIcon::SizeSmall),
        itemName,
        info.filePath());

      QFile file(info.filePath() + "/.directory");

      if(file.open(IO_ReadOnly))
      {
        item->desktopFile = QTextStream(&file).read();

        file.close();
      }

      push_back(parent.children, item);

      read_dir(info.filePath(), *item);
    }
    else
    {
      KDesktopFile desktopFile(info.filePath());
      QString itemName;

      if(desktopFile.readBoolEntry("NoDisplay") || desktopFile.readBoolEntry("Hidden"))
        continue;

      itemName = desktopFile.readName();
      if (!itemName)
        itemName = dir[i];
 
      MenuItem* item = new MenuItem(
        KGlobal::iconLoader()->loadIcon(desktopFile.readIcon(), KIcon::Desktop, KIcon::SizeSmall),
        itemName, info.filePath());

      push_back(parent.children, item);
    }
  }
}

KSchoolmenu::KSchoolmenu(QWidget* parent, const char* name)
  : KCModule(parent, name),
    availableItems(QPixmap(), QString("Root menu"), "/usr/share/applnk"),
    listAvailableItems(new QListView(this)),
    listCurrentMenu(new QListView(this)),
    comboCustomMenus(new QComboBox(this)),
    buttonAdd(new QPushButton(i18n("&Add..."), this)),
    buttonRemove(new QPushButton(i18n("&Remove"), this)),
    buttonRight(new QPushButton(SmallIcon("forward"), i18n("A&dd"), this)),
    buttonLeft(new QPushButton(SmallIcon("back"), i18n("R&emove"), this)),
    labelCustomMenu(new QLabel(comboCustomMenus, i18n("&Custom menu:"), this)),
    labelAvailableItems(new QLabel(listAvailableItems, i18n("A&vailable items:"), this)),
    labelCurrentMenu(new QLabel(listCurrentMenu, i18n("&Items in current menu:"), this))
{

  QHBoxLayout* layoutTop = new QHBoxLayout(0, 0, 6);
  layoutTop->addWidget(labelCustomMenu);
  layoutTop->addWidget(comboCustomMenus);
  layoutTop->addWidget(buttonAdd);
  layoutTop->addWidget(buttonRemove);
  layoutTop->addStretch();

  QVBoxLayout* layoutLeft = new QVBoxLayout(0, 0, 6);
  layoutLeft->addWidget(labelAvailableItems);
  layoutLeft->addWidget(listAvailableItems);

  QVBoxLayout* layoutCenter = new QVBoxLayout(0, 0, 6);
  layoutCenter->addStretch();
  layoutCenter->addWidget(buttonRight);
  layoutCenter->addWidget(buttonLeft);
  layoutCenter->addStretch();

  QVBoxLayout* layoutRight = new QVBoxLayout(0, 0, 6);
  layoutRight->addWidget(labelCurrentMenu);
  layoutRight->addWidget(listCurrentMenu);

  QHBoxLayout* layoutBottom = new QHBoxLayout(0, 0, 6);
  layoutBottom->addLayout(layoutLeft);
  layoutBottom->addLayout(layoutCenter);
  layoutBottom->addLayout(layoutRight);

  QVBoxLayout* layoutFull = new QVBoxLayout(this, 0, 6);
  layoutFull->addLayout(layoutTop);
  layoutFull->addLayout(layoutBottom);

  if(getuid() != 0)
    buttonAdd->setEnabled(false);

  buttonRight->setEnabled(false);
  buttonLeft->setEnabled(false);
  buttonRemove->setEnabled(false);

  connect(buttonAdd, SIGNAL(clicked()), this, SLOT(buttonAddClicked()));
  connect(buttonRemove, SIGNAL(clicked()), this, SLOT(buttonRemoveClicked()));
  connect(buttonRight, SIGNAL(clicked()), this, SLOT(buttonRightClicked()));
  connect(buttonLeft, SIGNAL(clicked()), this, SLOT(buttonLeftClicked()));
  connect(listAvailableItems, SIGNAL(selectionChanged()), this, SLOT(listAvailableItemsSelectionChanged()));
  connect(listCurrentMenu, SIGNAL(selectionChanged()), this, SLOT(listCurrentMenuSelectionChanged()));
  connect(comboCustomMenus, SIGNAL(activated(int)), this, SLOT(comboCustomMenusActivated(int)));

  load();
};

KSchoolmenu::~KSchoolmenu()
{
  // XXX: free memory
}

template<class type>
static void add_children(const MenuItem& parentItem, type* parentWidget, const QString &open, const QString &select)
{
  for(uint i = 0; i < parentItem.children.size(); ++i)
  {
    if (! parentItem.children[i]->Hidden())
    {
      QListViewItem* item = new QListViewItem(parentWidget, parentItem.children[i]->name, parentItem.children[i]->path);

      item->setPixmap(0, parentItem.children[i]->icon);

      if(!parentItem.children[i]->path.contains(".desktop"))
      {
        item->setExpandable(true);
	if (open.contains(parentItem.children[i]->path))
	  item->setOpen(true);
      }
      if (select == parentItem.children[i]->path)
        item->setSelected(true);

      add_children(*parentItem.children[i], item, open, select);
    }
  }
}

void KSchoolmenu::load()
{
  availableItems.children.clear();
  read_dir("/usr/share/applnk", availableItems);

  listAvailableItems->clear();
  listAvailableItems->addColumn("");
  listAvailableItems->header()->hide();
  listAvailableItems->setRootIsDecorated(true);

  comboCustomMenus->clear();

  QDir dir("/var/lib/kschoolmenu");
  dir.setSorting(QDir::Name);
  dir.setMatchAllDirs(true);

  // for every directory in /var/lib/kschoolmenu/
  for(uint i = 0; i < dir.count(); ++i)
  {
    if(dir[i] == "." || dir[i] == "..")
      continue;

    MenuItem* item = new MenuItem(QPixmap(), dir[i], QString("/var/lib/kschoolmenu/") + dir[i] + "/share/applnk");

    // read the files for the QListViews
    read_dir ("/usr/share/applnk", *item);
    // hide the files for the listCurrentMenu QListView
    read_desktop_files(item, "/var/lib/kschoolmenu/" + dir[i] + "/share") ;

    push_back(customMenus, item);

    comboCustomMenus->insertItem(dir[i]);
  }

  listCurrentMenu->clear();
  listCurrentMenu->addColumn("");
  listCurrentMenu->header()->hide();
  listCurrentMenu->setRootIsDecorated(true);

  add_children(availableItems, listAvailableItems, NULL, NULL);

  comboCustomMenusActivated(comboCustomMenus->currentItem());
}

MenuItem* find_hidden(MenuItem* menu)
{
  if(menu->Hidden())
    return menu;

  for(uint i = 0; i < menu->children.size(); ++i)
  {
    MenuItem* ret = find_hidden(menu->children[i]);

    if(ret)
      return ret;
  }

  return 0;
}

void clean(const QString& path)
{
  QDir dir;

  int p = 0;
  while(-1 != (p = path.find('/', p + 1)))
    dir.mkdir(path.left(p));

  dir.mkdir(path);

  dir.setPath(path);
  dir.setSorting(QDir::Name);
  dir.setFilter(QDir::Dirs | QDir::Files | QDir::Hidden);

  for(uint i = 0; i < dir.count(); ++i)
  {
    if(dir[i] == "." || dir[i] == "..")
      continue;

    QFileInfo info(dir, dir[i]);

    if(info.isDir())
    {
      clean(info.filePath());
      dir.rmdir(info.filePath());
    }
    else
    {
      dir.remove(dir[i]);
    }
  }
}

static void write_desktop_files(MenuItem* item, const QString& path)
{
  if(item->path.contains(".desktop"))
  {
    QFile file(path + item->path.right(item->path.length() - item->path.findRev('/')));
    if (item->Hidden())
    {
      item->desktopFile = "[Desktop Entry]\nHidden=true\nName=empty";

      if(file.open(IO_WriteOnly))
      {
        QTextStream stream(&file);
        stream << item->desktopFile;
        file.close();
      }
    }
    else
      file.remove(); 
  }
  else
  {
    QFile file(path + item->path.right(item->path.length() - item->path.findRev('/')) + "/.directory");
    if (find_hidden(item))
      QDir().mkdir(path + item->path.right(item->path.length() - item->path.findRev('/')));
    if(item->Hidden())
    {
      item->desktopFile = "[Desktop Entry]\nHidden=true\nName=empty";

      if(file.open(IO_WriteOnly))
      {
        QTextStream stream(&file);
        stream << item->desktopFile;
        file.close();
      }
    }
    else
      file.remove(); 

    for(uint i = 0; i < item->children.size(); ++i)
      if (path == "")
        write_desktop_files(item->children[i], path + item->path);
      else
        write_desktop_files(item->children[i], path + '/' + item->path.right(item->path.length() - item->path.findRev('/') - 1));
  }
}

void KSchoolmenu::save()
{
  QDir dir("/var/lib/kschoolmenu");
  dir.setSorting(QDir::Name);
  dir.setMatchAllDirs(true);
  
  for(uint i = 0; i < dir.count(); ++i)
  {
    if(dir[i] == "." || dir[i] == "..")
      continue;

    QFileInfo info(dir, dir[i]);


    if(info.isDir())
    {
      clean(info.filePath());


      QDir().rmdir(info.filePath());
    }
  }

  for(uint i = 0; i < customMenus.size(); ++i)
  {
    QString path = QString("/var/lib/kschoolmenu/") + customMenus[i]->name;

    dir.mkdir(path);
    path += "/share";
    dir.mkdir(path);
    path += "/applnk";
    dir.mkdir(path);

    write_desktop_files(customMenus[i], "");
  }

  emit changed(true);
}

int KSchoolmenu::buttons()
{
  return KCModule::Apply;
}

QString KSchoolmenu::quickHelp() const
{
  return i18n("Configure custom K menus."); 
}

void KSchoolmenu::buttonAddClicked()
{
  bool ok;

  QString name = KLineEditDlg::getText(i18n("Name of new menu"), i18n("Untitled"), &ok, this);

  if(!ok)
    return;

  for(uint i = 0; i < this->customMenus.size(); ++i)
  {
    if(customMenus[i]->name == name)
    {
      KMessageBox::error(this, i18n("The selected name is already in use by another menu."));

      return;
    }
  }

 
  MenuItem* item = new MenuItem(QPixmap(), name, QString("/var/lib/kschoolmenu/") + name + "/share/applnk");

  read_dir ("/usr/share/applnk", *item);

  push_back(customMenus, item);


  comboCustomMenus->insertItem(name);

  if(comboCustomMenus->currentItem() == -1)
    comboCustomMenus->setCurrentItem(0);

  comboCustomMenusActivated(comboCustomMenus->currentItem());

  emit changed(true);
}
  
void KSchoolmenu::buttonRemoveClicked()
{
  if(comboCustomMenus->currentText() == "template")
  {
    KMessageBox::error(this, i18n("You cannot remove the \"template\" menu."));

    return;
  }
  
  for(uint i = 0; i < customMenus.size(); ++i)
  {
    if(customMenus[i]->name == comboCustomMenus->currentText())
    {
      comboCustomMenus->removeItem(comboCustomMenus->currentItem());

      if(customMenus.size() == 1)
      {
        customMenus.resize(0);

        comboCustomMenusActivated(-1);

        break;
      }

      if(i != customMenus.size() - 1)
        customMenus.insert(i, customMenus.take(customMenus.size() - 1));

      customMenus.resize(customMenus.size() - 1);

      break;
    }
  }

  emit changed(true);
}

MenuItem* find_item(MenuItem* menu, const QString& path, const QString& name)
{
  if(menu->path.left(menu->path.findRev('/')) == path)
  {
    if(menu->path.contains(".desktop"))
    { 
      KDesktopFile desktopFile(menu->path);

      if(desktopFile.readName() == name)
        return menu;
    }
    else
    {
      KDesktopFile desktopFile(menu->path + "/.directory");

      if(desktopFile.readName() == name || menu->path.right(menu->path.length() - menu->path.findRev('/') - 1) == name)
        return menu;
    } 
  }

  for(uint i = 0; i < menu->children.size(); ++i)
  {
    MenuItem* ret = find_item(menu->children[i], path, name);

    if(ret)
      return ret;
  }

  return 0;
}

void KSchoolmenu::buttonRightClicked()
{
  QListViewItem* item = listAvailableItems->currentItem();

  MenuItem* menu;

  for(uint i = 0; i < customMenus.size(); ++i)
    if(customMenus[i]->name == comboCustomMenus->currentText())
    {
      menu = customMenus[i];

      break;
    }

  MenuItem *tmpItem = find_item(menu, item->text(1).left(item->text(1).findRev('/')), item->text(0));
  QString open = item->text(1);
  QString select = open;
  if (tmpItem->Hidden())
  {
    tmpItem->Show();
    listCurrentMenu->clear();
    add_children(*menu, listCurrentMenu, open, select);
  }

  while (item->parent())
  {
    item = item->parent(); 
    open += item->text(1);
    MenuItem *tmpItem = find_item(menu, item->text(1).left(item->text(1).findRev('/')), item->text(0));
    if (tmpItem->Hidden())
    {
      tmpItem->Show();
      listCurrentMenu->clear();
      add_children(*menu, listCurrentMenu, open, select);
    }
  }
  
  emit changed(true);
}

void KSchoolmenu::buttonLeftClicked()
{
  QListViewItem* toHide = listCurrentMenu->currentItem() ;

  MenuItem* item;

  // Search to find CustomMenu we are working with
  for(uint i = 0; i < customMenus.size(); ++i)
    if(customMenus[i]->name == comboCustomMenus->currentText())
    {
      item = customMenus[i];

      break;
    }

  // search to find top List-handle 
  QListViewItem* topFound = toHide ;
  QString path = "/usr/share/applnk"; 
 
  while (topFound->parent()) 
    topFound = topFound->parent() ;
  
  if (toHide->parent())
    path = toHide->parent()->text(1);
  
  while (topFound != toHide)
  {
    item = find_item (item, path, toHide->text(0));
    QListViewItem* listTraverse = toHide ; 
    while (listTraverse->parent() != topFound)
      listTraverse = listTraverse->parent();
    topFound = listTraverse ; 
  }

  if (topFound == toHide)
  {
    item = find_item (item, path, toHide->text(0));
    item->Hide() ; 
    delete listCurrentMenu->currentItem();

    if(listCurrentMenu->currentItem())
      listCurrentMenu->setSelected(listCurrentMenu->currentItem(), true);
  }

  emit changed(true);
}

void KSchoolmenu::listAvailableItemsSelectionChanged()
{

  if(!listAvailableItems->currentItem())
  {
    buttonRight->setEnabled(false);

    return;
  }

  // if(!listAvailableItems->currentItem()->isExpandable() && buttonRemove->isEnabled())
  if(buttonRemove->isEnabled())
  {
    buttonRight->setEnabled(true);
  }
  else
  {
    buttonRight->setEnabled(false);
  }
}

void KSchoolmenu::listCurrentMenuSelectionChanged()
{
  if(!listCurrentMenu->currentItem())
  {
    buttonLeft->setEnabled(false);

    return;
  }

  // if(!listCurrentMenu->currentItem()->isExpandable() && buttonRemove->isEnabled())
  if(buttonRemove->isEnabled())
  {
    buttonLeft->setEnabled(true);
  }
  else
  {
    buttonLeft->setEnabled(false);
  }
}


void KSchoolmenu::comboCustomMenusActivated(int index)
{
  if(index >= 0 && index < comboCustomMenus->count())
  {
    listCurrentMenu->clear();

    for(uint i = 0; i < customMenus.size(); ++i)
    {
      if(customMenus[i]->name == comboCustomMenus->currentText())
      {
        add_children(*customMenus[i], listCurrentMenu, NULL, NULL);

        break;
      }
    }
    
    if(getuid() == 0)
    {
      buttonRemove->setEnabled(true);
      listAvailableItemsSelectionChanged();
      listCurrentMenuSelectionChanged();
    }
  }
  else
  {
    listCurrentMenu->clear();
    buttonRemove->setEnabled(false);
    buttonLeft->setEnabled(false);
    buttonRight->setEnabled(false);
  }
}


// ------------------------------------------------------------------------

extern "C"
{
  KCModule* create_kschoolmenu(QWidget* parent, const char* name)
  {
    KGlobal::locale()->insertCatalogue("kschoolmenu");

    return new KSchoolmenu(parent, name);
  }
}

#include "kschoolmenu.moc"
